#include "besupport/MessageTransceiverThread.h"
#include "reflector/StorageReflectConstants.h"
#include <View.h>
#include <Window.h>
#include <Alert.h>
#include <Application.h>
#include <StringView.h>
#include <ListView.h>
#include <ListItem.h>
#include <ScrollView.h>
#include <TextControl.h>
#include <Node.h>
#include <Roster.h>
#include <Path.h>
#include <TranslationUtils.h>
#include <Bitmap.h>

bool our_turn = false;

typedef enum {
	none = -1,
	destroyer,	//----2 spaces
	submarine,	//----3 spaces
	cruiser, 	//----3 spaces
	battleship,	//----4 spaces
	carrier		//----5 spaces
} ship_type;

const uint8 ship_length[] = {2,3,3,4,5};
const char ship_signal[] = {'D','S','C','B','A'};
const char *ship_name[] = {"destroyer","submarine","cruiser","battleship","carrier"};
char key[255];
char us[255];
char our_key[255];

enum {
	HIT_SQUARE,
	HIT_RESULT,
	SHIP_SUNK,
	PLAY_BALL,
	YOUR_TURN,
	HI,
	HI_REJECT,
	YOU_LOSE
};

#define window ((GameBoard *)(Window()))

BBitmap *get_bitmap(int orient);

class KeyItem : public BStringItem {
	public:
		KeyItem(PortableString &string) : BStringItem(nameFromKey(string)) {
			strcpy(key,string.Cstr());
		}
		const char *Key(void) {
			return key;
		}
	private:
		const char *nameFromKey(PortableString &string) {	
			int32 last = 0;
			last = string.IndexOf("/name/") + 6;
			const char *name = strdup(string.Substring(last).Cstr());
			printf("Name is %s, key is %s, last is %d\n",name,string.Cstr(),last);
			return name;
		}
		char key[255];
};

class square : public BView {
	public:
		square(int board,int x, int y) : BView(BRect(x*20,y*20,(x*20)+20,(y*20)+20),NULL,B_FOLLOW_LEFT | B_FOLLOW_TOP,B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE) {
			SetViewColor(98,109,188);
			SetLowColor(98,109,188);
			ship = none;
			this->x = x;
			this->y = y;
			board_type = board;
			hit = 0;
			place = 0;
			pos = 0;
		}
		void SetShip(ship_type ship, int position, int direction) {
			this->ship = ship;
			place = position;
			pos = direction;
		}
		void Hit();
		virtual	void			Draw(BRect updateRect) {
			//--picts c/o Jason Matthews
			SetLowColor(ViewColor());
			if (ship != none) {
				int offset = ((place == (ship_length[ship] - 1)) ? (21*4) : (21*place));
				BRect src_rect = get_bitmap(pos)->Bounds();
				switch (pos) {
					case 0:
						src_rect.left += offset;
						src_rect.right = src_rect.left + 18;
						break;
					case 1:
						src_rect.top += offset;
						src_rect.bottom = src_rect.top + 18;
						break;
					case 2:
						src_rect.right -= offset;
						src_rect.left = src_rect.right - 18;
						break;
					case 3:
						src_rect.bottom -= offset;
						src_rect.top = src_rect.bottom - 18;
						break;
				}				
				BRect dest = Bounds();
				dest.InsetBy(1,1);
				printf("square # = %d, ",place);
				printf("src_rect = (%.2f,%.2f",src_rect.left,src_rect.top);
				printf(",%.2f,%.2f)",src_rect.right,src_rect.bottom);
				printf(", offset = %d, ship = %d (%s), orientation = %d\n",offset,ship,ship_name[ship],pos);
				DrawBitmap(get_bitmap(pos),src_rect,dest);
			}
			if (hit == 1)
				SetHighColor(255,0,0);
			if (hit == -1)
				SetHighColor(0,255,0);
			if (hit != 0)
				FillArc(BRect(7,7,13,13),0,360);
			SetHighColor(0,0,0);
			StrokeRect(Bounds());
		}
		virtual	void MouseDown(BPoint where);
	private:
		friend class GameBoard;
		int place, pos;
		int x,y;
		int board_type;
		ship_type ship;
		int hit;
};

class GameBoard : public BWindow {
	public:
		GameBoard(const char *IP = "beshare.befaqs.com") : BWindow(BRect(200,100,560,600),"Battleship",B_TITLED_WINDOW,B_NOT_RESIZABLE | B_NOT_ZOOMABLE) {
			lastx = -1;
			lasty = -1;
			placing = none;
			partner_ready = false;
			connected = false;
			BView *opponent, *me;
			opponent = new BView(BRect(25,25,225,225),"opponent",B_FOLLOW_LEFT | B_FOLLOW_TOP,0);
			me = new BView(BRect(25,250,225,450),"me",B_FOLLOW_LEFT | B_FOLLOW_TOP,0);
			mtt = new MessageTransceiverThread(BMessenger(this));
			name_box = new BTextControl(BRect(230,25,340,45),"name_box","Name: ",us,new BMessage('name'));
			name_box->SetDivider(be_plain_font->StringWidth("Name: "));
			AddChild(name_box);
			players = new BListView(BRect(230,50,320,448),"playerslist");
			playersbox = new BScrollView("playersbox",players,B_FOLLOW_LEFT | B_FOLLOW_TOP, 0, false, true);
//			players->AddItem(new BStringView("No players"));
			players->SetInvocationMessage(new BMessage('chos'));
			AddChild(playersbox);
			for (uint8 x = 0; x < 10; x++) {
				for (uint8 y = 0; y < 10; y++) {
					squares[0][x][y] = new square(1,x,y);
					me->AddChild(squares[0][x][y]);
				}
			}
			for (uint8 x = 0; x < 10; x++) {
				for (uint8 y = 0; y < 10; y++) {
					squares[1][x][y] = new square(2,x,y);
					opponent->AddChild(squares[1][x][y]);
				}
			}
			for (int i = 0; i < 5; i++)
				ship_count[i] = 0;
			status = new BStringView(BRect(5,230,245,245),"status","Waiting for opponent to connect...");
			AddChild(status);
			AddChild(opponent);
			AddChild(me);
			Show();
			accept = false;
			mtt->StartConnectThread(IP,2960,new BMessage('conn'),new BMessage('disc'));
			ships_sunk = 0;
		}
		void Hit(ship_type ship) {
			puts("Ship Hit (GameBoard)! Waahoo!");
			ship_count[ship]++;
			if (ship_count[ship] >= ship_length[ship]) {
				printf("Sending SHIP_SUNK: ship_count for %s is %d, ship_length is %d\n",ship_name[ship],ship_count[ship],ship_length[ship]);
				PortableMessage *msg = new PortableMessage(SHIP_SUNK);
				msg->AddInt8("ship",ship);
				msg->AddString(PR_NAME_KEYS, key);
				mtt->SendOutgoingMessage(PortableMessageRef(msg, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
			}
		}
		void MessageReceived(BMessage *msg) {
			PortableMessage *msg2;
			switch (msg->what) {
				case PORTABLE_MESSAGES_RECEIVED:
				{
					puts("Portable message received! Waahoo!");
					PortableMessageRef next;
					PortableMessage *msg;
	       			int8 x,y;
	       			bool result;
	         		while (mtt->GetNextIncomingMessage(next) == B_NO_ERROR) {
	         			msg = next.GetItemPointer();
	         			printf("msg received with what: %d:\n",msg->what);
	         			msg->PrintToStream();
	         			switch (msg->what) {
	         				case HIT_SQUARE:
	         					msg->FindInt8("x",0,&x);
	         					msg->FindInt8("y",0,&y);
	         					squares[0][x][y]->Hit();
	         					break;
	         				case HIT_RESULT:
	         					msg->FindInt8("x",0,&x);
	         					msg->FindInt8("y",0,&y);
	         					msg->FindBool("result",0,&result);
	         					squares[1][x][y]->hit = (result) ? 1 : -1;
	         					squares[1][x][y]->Invalidate();
	         					msg2 = new PortableMessage(YOUR_TURN);
	         					msg2->AddString(PR_NAME_KEYS, key);
								mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
								status->SetText("Duck and Cover!");
	         					break;
	         				case SHIP_SUNK:
	         					{
	         						char message[255];
	         						msg->FindInt8("ship",0,&x);
	         						sprintf(message,"You have sunk the enemy's %s!",ship_name[x]);
	         						(new BAlert("sunk",message,"Yay!"))->Go();
	         						ships_sunk++;
	         						if (ships_sunk >= 5) {
	         							msg2 = new PortableMessage(YOU_LOSE);
	         							msg2->AddString(PR_NAME_KEYS,key);
	         							mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
	         							(new BAlert("win","You win!","Hooray!"))->Go();
	         							key[0] = 0;
			         					our_turn = false;
			         					connected = false;
										AddChild(playersbox);
										AddChild(name_box);
			         					ResizeTo(360,500);
			         					lastx = -1;
			         					lasty = -1;
			         					status->SetText("Waiting for opponent to connect...");
	         						}
	         					}
	         					break;
	         				case YOU_LOSE:
	         					(new BAlert("win","You lose!","Phooey"))->Go();
	         					key[0] = 0;
	         					our_turn = false;
	         					connected = false;
								AddChild(playersbox);
								AddChild(name_box);
	         					ResizeTo(360,500);
	         					lastx = -1;
	         					lasty = -1;
			         			status->SetText("Waiting for opponent to connect...");
	         					break;
	         				case PLAY_BALL:
	         					partner_ready = true;
	         					if (accept)
	         						our_turn = true;
	         					if ((placing == none) && (!accept))
	         						status->SetText("Duck and Cover!");
	         					if (!((placing == none) && (accept)))
	         						break;
	         				case YOUR_TURN:
	         					our_turn = true;
	         					status->SetText("Fire at will!");
	         					break;
	         				case PR_RESULT_DATAITEMS:
	         					{
	         						PortableString name;
	         						type_code bs;
	         						KeyItem *item;
	         						for (int32 i = 0; msg->GetInfo(B_MESSAGE_TYPE,i,name,&bs) == B_OK; i++)
	         							players->AddItem(new KeyItem(name));
	         						for (int32 i = 0; msg->FindString(PR_NAME_REMOVED_DATAITEMS,i,name) == B_OK; i++) {
	         							for (int32 w = 0; (item = (KeyItem *)(players->ItemAt(w))) != NULL; w++) {
	         								if (name == item->Key()) {
	         									players->RemoveItem(item);
	         									break;
	         								}
	         							}
	         						}
	         					}
	         					break;
	         				case HI_REJECT:
	         					(new BAlert("dang","That user already playing a game.","OK"))->Go();
	         					break;
	         				case HI:
	         					{
	         						bool first = false;
	         						if (msg->FindBool("am_first",&first) == B_OK)
	         							accept = true;
									const char *them;
									msg->FindString("session",&them);
									if (!connected) {
										char text[255], *flip;
										for (uint8 x = 0; x < 10; x++) {
											for (uint8 y = 0; y < 10; y++) {
												squares[0][x][y]->SetShip(none,0,0);
												squares[0][x][y]->hit = 0;
												squares[0][x][y]->SetHighColor(squares[0][x][y]->ViewColor());
												squares[0][x][y]->FillRect(squares[0][x][y]->Bounds());
												squares[0][x][y]->Draw(squares[0][x][y]->Bounds());
											}
										}
										for (uint8 x = 0; x < 10; x++) {
											for (uint8 y = 0; y < 10; y++) {
												squares[1][x][y]->SetShip(none,0,0);
												squares[1][x][y]->hit = 0;
												squares[1][x][y]->SetHighColor(squares[1][x][y]->ViewColor());
												squares[1][x][y]->FillRect(squares[1][x][y]->Bounds());
												squares[1][x][y]->Draw(squares[1][x][y]->Bounds());
											}
										}
										msg->FindString("user",(const char **)&flip);
										sprintf(text,"%s Connected. Please place destroyer...",flip);
										status->SetText(text);
										puts("Connected! Yes!");
										sprintf(key,"/*/%s",them);
										puts("Key is as follows:");
										puts(key);
										placing = destroyer;
										msg2 = new PortableMessage(HI);
										msg2->AddString("session",our_key);
	         							msg2->AddString("user",us);
			    						msg2->AddString(PR_NAME_KEYS, key);
										mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
										msg2 = new PortableMessage(PR_COMMAND_SETDATA);
		         						PortableMessage name;
		         						name.AddBool("is_free",false);
		         						msg2->AddMessage("battleship/is_free",name);
		    							puts("Sending message:");
		    							
		    							ship_count[0] = 0;
		    							ship_count[1] = 0;
		    							ship_count[2] = 0;
		    							ship_count[3] = 0;
		    							ship_count[4] = 0;
		    							
		    							ships_sunk = 0;
		    							
		         						mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
									} else if (first) {
										char blat[255];
										sprintf(blat,"/*/%s",them);
										msg2 = new PortableMessage(HI_REJECT);
			    						msg2->AddString(PR_NAME_KEYS, blat);
										mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
									}
									ResizeTo(250,500);
									RemoveChild(playersbox);
									RemoveChild(name_box);
									connected = true;
									break;
	         					}
							case PR_RESULT_PARAMETERS:
								{
							         const char * sessionRoot;
							         if (msg->FindString(PR_NAME_SESSION_ROOT, &sessionRoot) == B_NO_ERROR)
							         {
							            // session root is of form "/hostname/sessionID"; parse these out
							            const char * lastSlash = strrchr(sessionRoot, '/');
							            if (lastSlash)
							            {
							               sprintf(our_key,"%s/battleship",lastSlash+1);
							            }
							         }
							      }
							      break;
		         			}
		         		}
		         	}
		         	break;
		        case 'chos':
		        	if (players->CurrentSelection() < 0)
		        		break;
		        	strcpy(key,((KeyItem *)(players->ItemAt(players->CurrentSelection())))->Key());
		        	msg2 = new PortableMessage(HI);
	         		msg2->AddBool("am_first",true);
	         		msg2->AddString("session",our_key);
	         		msg2->AddString("user",us);
	    			msg2->AddString(PR_NAME_KEYS, key);
	    			puts("Sending message:");
	    			msg2->PrintToStream();
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
	         		break;
	         	case 'conn':
				{
					PortableMessage name;
					char text[255];
					sprintf(text,"battleship/name/%s",us);
					msg2 = new PortableMessage(PR_COMMAND_SETDATA);
					msg2->AddMessage(text,name);
					name.AddBool("is_free",true);
					msg2->AddMessage("battleship/is_free",name);
	    			puts("Sending message:");
	    			msg2->PrintToStream();
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					msg2 = new PortableMessage(PR_COMMAND_SETPARAMETERS);
					msg2->AddBool("SUBSCRIBE:/*/*/battleship/name/*",true);
	    			puts("Sending message:");
	    			msg2->PrintToStream();
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					msg2 = new PortableMessage(PR_COMMAND_GETPARAMETERS);
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					break;
				}
	         	case 'disc':
					(new BAlert("eek","You have been disconnected from the server.","OK"));
					be_app->PostMessage(B_QUIT_REQUESTED);
					break;
				case 'name':
					msg2 = new PortableMessage(PR_COMMAND_REMOVEDATA);
					msg2->AddString(PR_NAME_KEYS, "battleship/name");
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					strcpy(us,name_box->Text());
					PortableMessage name;
					char text[255];
					sprintf(text,"battleship/name/%s",us);
					msg2 = new PortableMessage(PR_COMMAND_SETDATA);
					msg2->AddMessage(text,name);
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					msg2 = new PortableMessage(PR_COMMAND_GETPARAMETERS);
					mtt->SendOutgoingMessage(PortableMessageRef(msg2, mtt->GetMessagePool(), mtt->GetMessageRefPool()));
					app_info info;
					be_app->GetAppInfo(&info);
					BNode(&(info.ref)).WriteAttr("user_name",'CSTR',0,us,strlen(us)+1);
					break;
			}
	     }
         bool QuitRequested(void) {
         	be_app->PostMessage(B_QUIT_REQUESTED);
         	return true;
         }
         
         ship_type placing;
         int lastx, lasty;
	private:
		bool partner_ready;
		bool accept;
		bool connected;
		MessageTransceiverThread *mtt;
		BStringView *status;
		BTextControl *name_box;
		BListView *players;
		BScrollView *playersbox;
		square *squares[2][10][10];
		int ship_count[5];
		int ships_sunk;
		friend class square;
};


void square::MouseDown(BPoint where) {
	puts("Mousing! Waahoo!");
	if ((board_type == 2) && (hit == 0) && (our_turn == true)) {
		puts("Hitting! Waahoo!");
		PortableMessage *msg = new PortableMessage(HIT_SQUARE);
		msg->AddInt8("x",x);
		msg->AddInt8("y",y);
		msg->AddString(PR_NAME_KEYS, key);
		((GameBoard *)(Window()))->mtt->SendOutgoingMessage(PortableMessageRef(msg, ((GameBoard *)(Window()))->mtt->GetMessagePool(), ((GameBoard *)(Window()))->mtt->GetMessageRefPool()));
		our_turn = false;
	} else if ((((GameBoard *)(Window()))->placing != none) && (board_type == 1)) {
		if (ship != none)
			return;
		
		if (((GameBoard *)(Window()))->lastx == -1) {
			((GameBoard *)(Window()))->lastx = x;
			((GameBoard *)(Window()))->lasty = y;
			ship = ((GameBoard *)(Window()))->placing;
			Invalidate();
		} else {
			if (abs(x-((GameBoard *)(Window()))->lastx) > abs(y-((GameBoard *)(Window()))->lasty)) {
				int inc = (x-((GameBoard *)(Window()))->lastx > 0) ? 1 : -1;
				int temp_lastx = window->lastx;
				if (window->lastx + (inc*ship_length[window->placing]) < 0)
					temp_lastx = ship_length[window->placing];
				if (window->lastx + (inc*ship_length[window->placing]) > 10)
					temp_lastx = 10 - ship_length[window->placing];
				int workx = temp_lastx+inc;
				for (;abs(workx-window->lastx) < ship_length[window->placing]; workx += inc) {
					if (window->squares[0][workx][window->lasty]->ship != none)
						return;
				}
				window->lastx = temp_lastx;
				workx = window->lastx;
				for (;abs(workx-window->lastx) < ship_length[window->placing]; workx += inc) {
					window->squares[0][workx][window->lasty]->SetShip(window->placing,abs(workx-window->lastx),(inc == -1) ? 2 : 0);
					window->squares[0][workx][window->lasty]->Invalidate();
				}
			} else {
				int inc = (y-((GameBoard *)(Window()))->lasty > 0) ? 1 : -1;
				int temp_lasty = window->lasty;
				if (window->lasty + (inc*ship_length[window->placing]) < 0)
					temp_lasty = ship_length[window->placing];
				if (window->lasty + (inc*ship_length[window->placing]) > 10)
					temp_lasty = 10 - ship_length[window->placing];
				int worky = temp_lasty+inc;
				for (;abs(worky-window->lasty) < ship_length[window->placing]; worky += inc) {
					if (window->squares[0][window->lastx][worky]->ship != none)
						return;
				}
				window->lasty = temp_lasty;
				worky = window->lasty;
				for (;abs(worky-window->lasty) < ship_length[window->placing]; worky += inc) {
					window->squares[0][window->lastx][worky]->SetShip(window->placing,abs(worky-window->lasty),(inc == -1) ? 3 : 1);
					window->squares[0][window->lastx][worky]->Invalidate();
				}
				
			}
			if (window->placing < carrier) {
				window->placing = ship_type(int(window->placing) + 1);
				char message[255];
				window->lastx = -1;
				window->lasty = -1;
				sprintf(message,"Please place %s...",ship_name[window->placing]);
				window->status->SetText(message);
			} else {
				window->placing = none;
				PortableMessage *msg = new PortableMessage(PLAY_BALL);
				msg->AddString(PR_NAME_KEYS, key);
				window->mtt->SendOutgoingMessage(PortableMessageRef(msg, window->mtt->GetMessagePool(), window->mtt->GetMessageRefPool()));
				window->status->SetText((our_turn) ? "Fire at will!" : "Waiting for opponent to finish placing ships...");
			}
		}
			
	}
}

void square::Hit() {
	if (board_type == 1) {
		puts("We're hit! Waahoo!");
		hit = (ship != none) ? 1 : -1;
		PortableMessage *msg = new PortableMessage(HIT_RESULT);
		msg->AddInt8("x",x);
		msg->AddInt8("y",y);
		msg->AddBool("result",(hit == 1));
		msg->AddString(PR_NAME_KEYS, key);
		((GameBoard *)(Window()))->mtt->SendOutgoingMessage(PortableMessageRef(msg, ((GameBoard *)(Window()))->mtt->GetMessagePool(), ((GameBoard *)(Window()))->mtt->GetMessageRefPool()));
		if (ship != none)
			((GameBoard *)(Window()))->Hit(ship);
	}
	Invalidate();
}

BBitmap *get_bitmap(int orient) {
	static BBitmap *picts[4] = {NULL,NULL,NULL,NULL};
	static BPath dir;
	char *names[4] = {"right","down","left","up"};
	if (picts[orient] == NULL) {
		if (dir.InitCheck() != B_OK) {
			app_info info;
			be_app->GetAppInfo(&info);
			BPath path(&(info.ref));
			path.GetParent(&dir);
			dir.Append("img");
		}
		BPath file(dir.Path(),names[orient]);
		puts(file.Path());
		picts[orient] = BTranslationUtils::GetBitmapFile(file.Path());
		if (picts[orient] == NULL)
			(new BAlert("ck","picts[orient] == NULL","OK"))->Go();
	}
	return picts[orient];
}

int main(int argc,const char **argv) {
	app_info info;
	be_app->GetAppInfo(&info);
	if (BNode(&(info.ref)).ReadAttr("user_name",'CSTR',0,us,255) <= 0)
		strcpy(us,"Anonymous");
	key[0] = 0;
	new BApplication("application/muscle-battleship");
	if (argc > 1)
		(new GameBoard(argv[1]));
	else
		(new GameBoard());
	be_app->Run();
	delete be_app;
}